/*
 * SPDX-License-Identifier: LGPL-3.0-or-later
 * Copyright (C) 2020 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <libgen.h>
#include <getopt.h>
#include <errno.h>
#include <fmc/core.h>
#include <linux/ipmi/fru.h>

static const char git_version[] = "git version: " GIT_VERSION;
static char *name;
static unsigned int verbose = 0;

static void help(void)
{
	fprintf(stderr, "%s: Lists FMC carriers and Mezzanines\n"
		"\t-V  print version\n"
		"\t-h  print help\n"
		"\t-v  verbose\n",
		name);
}

static void print_version(void)
{
	printf("%s version %s\nFMC library version: %s\n",
	       name, git_version, fmc_version_get());
}

static void print_carrier_name(struct fmc_tkn *tkn)
{
	char carr_name[32];

        fmc_carrier_name(tkn, carr_name, sizeof(carr_name));
	fprintf(stdout, "\tcarrier: %s\n", carr_name);
}

static void print_slot_name(struct fmc_tkn *tkn, unsigned int id)
{
	char *fru;
	int ret;

	fru = malloc(FRU_SIZE_MAX);
	if (!fru)
		goto err_alloc;

        ret = fmc_slot_eeprom_read(tkn, id, fru, FRU_SIZE_MAX, 0);
	if (ret < 0)
		goto err_read;
	if (ret == FRU_SIZE_MAX) {
		char *fmc_name;

                fmc_name = fru_get_product_name((struct fru_common_header *)fru);
		fprintf(stdout, "\t\tmodule: %s\n", fmc_name);
		free(fmc_name);
	}

	free(fru);
	return;

err_read:
	free(fru);
err_alloc:
        fprintf(stdout, "\t\tmodule: unknown (%s)\n",
		fmc_strerror(errno));
}

static void print_slot_flags(struct fmc_tkn *tkn, unsigned int id)
{
	fputs("\t\tflags:\n", stdout);
	fprintf(stdout,
		"\t\t\tpresent: %d\n",
		fmc_slot_is_present(tkn, id));
	fprintf(stdout,
		"\t\t\tfru-valid: %d\n",
		fmc_slot_is_fru_valid(tkn, id));
}

static void print_slot_geo_addr(struct fmc_tkn *tkn, unsigned int id)
{
	unsigned int ga;
	int err;

	err = fmc_slot_geo_address(tkn, id, &ga);
	if (err)
		fputs("\t\tgeo-addr: 0x--,\n", stdout);
	else
		fprintf(stdout, "\t\tgeo-addr: 0x%02x\n", ga);
}

#define BUF_LEN 16
static void print_slot_eeprom_type(struct fmc_tkn *tkn, unsigned int id)
{
	char buf[BUF_LEN];
	int err;

	err = fmc_slot_eeprom_type_get(tkn, id, buf, BUF_LEN);
	if (err)
		fputs("\t\teeprom-type: unknown,\n", stdout);
	else
		fprintf(stdout, "\t\teeprom-type: %s\n", buf);
}
#undef BUFLEN

static void print_slot(struct fmc_tkn *tkn,
		       unsigned int carr_id, unsigned int id)
{
	fprintf(stdout, "\t- fmc-slot-%u.%u:\n", carr_id, id);
	print_slot_name(tkn, id);
	if (verbose > 0)
		print_slot_flags(tkn, id);
	if (verbose > 1) {
		print_slot_geo_addr(tkn, id);
		print_slot_eeprom_type(tkn, id);
	}
}

int main(int argc, char *argv[])
{
	unsigned int *carr_id;
	int err = 0;
	int i;
	char opt;

	name = strndup(basename(argv[0]), 64);
	if (!name) {
		err = -1;
		goto out_strdup;
	}
	while ((opt = getopt(argc, argv, "h?Vv")) != -1) {
		switch (opt) {
		case 'h':
		case '?':
			help();
			exit(EXIT_SUCCESS);
			break;
		case 'V':
			print_version();
			exit(EXIT_SUCCESS);
		case 'v':
			verbose++;
			break;
		}
	}

	carr_id = fmc_carrier_id_list_get();
	if (!carr_id)
		goto out_list;
	for (i = 0; carr_id[i] != FMC_ID_INVALID; ++i) {
		struct fmc_tkn *tkn = fmc_carrier_open(carr_id[i]);
		unsigned int *slot_id;
		int k;

                fprintf(stdout, "fmc-carrier-%u:\n", carr_id[i]);
		print_carrier_name(tkn);
		slot_id = fmc_slot_id_list_get(tkn);
		if (!slot_id) {
			fputs("\tCan't get slot list\n", stderr);
			continue;
		}
		for (k = 0; slot_id[k] != FMC_ID_INVALID; ++k)
			print_slot(tkn, carr_id[i], slot_id[k]);
		fmc_slot_id_list_put(slot_id);
		fmc_carrier_close(tkn);
	}
	fmc_carrier_id_list_put(carr_id);
out_list:
	free(name);
out_strdup:
	exit(err ? EXIT_FAILURE : EXIT_SUCCESS);
}
